"""Process Manager Application."""

import asyncio
import logging
from typing import Dict, Optional

from astoria.common.broadcast_event import UsercodeLogBroadcastEvent
from astoria.common.manager import StateManager
from astoria.common.manager_requests import (
    RequestResponse,
    UsercodeKillManagerRequest,
    UsercodeRestartManagerRequest,
)
from astoria.common.messages.astdiskd import DiskInfo, DiskType, DiskUUID
from astoria.common.messages.astprocd import CodeStatus, ProcessManagerMessage
from astoria.common.mqtt import BroadcastHelper
from astoria.managers.mixins.disk_handler import DiskHandlerMixin

from .usercode_lifecycle import UsercodeLifecycle

LOGGER = logging.getLogger(__name__)

loop = asyncio.get_event_loop()


class ProcessManager(DiskHandlerMixin, StateManager[ProcessManagerMessage]):
    """Astoria Process State Manager."""

    name = "astprocd"
    dependencies = ["astdiskd", "astmetad"]

    def _init(self) -> None:
        self._lifecycle: Optional[UsercodeLifecycle] = None
        self._cur_disks: Dict[DiskUUID, DiskInfo] = {}

        self._mqtt.subscribe("astdiskd", self.handle_astdiskd_disk_info_message)

        self._register_request(
            "restart",
            UsercodeRestartManagerRequest,
            self.handle_restart_request,
        )
        self._register_request(
            "kill",
            UsercodeKillManagerRequest,
            self.handle_kill_request,
        )
        self._log_helper = BroadcastHelper.get_helper(
            self._mqtt,
            UsercodeLogBroadcastEvent,
        )

    @property
    def offline_status(self) -> ProcessManagerMessage:
        """
        Status to publish when the manager goes offline.

        This status should ensure that any other components relying
        on this data go into a safe state.
        """
        return ProcessManagerMessage(
            status=ProcessManagerMessage.Status.STOPPED,
        )

    async def main(self) -> None:
        """Main routine for astprocd."""
        # Wait whilst the program is running.
        self.update_status()
        await self.wait_loop()

        for uuid, info in self._cur_disks.items():
            asyncio.ensure_future(self.handle_disk_removal(uuid, info))

    async def handle_disk_insertion(self, uuid: DiskUUID, disk_info: DiskInfo) -> None:
        """Handle a disk insertion."""
        LOGGER.debug(f"Disk inserted: {uuid} ({disk_info.disk_type})")
        if disk_info.disk_type is DiskType.USERCODE:
            LOGGER.info(f"Usercode disk {uuid} is mounted at {disk_info.mount_path}")
            if self._lifecycle is None:
                LOGGER.debug(f"Starting usercode lifecycle for {uuid}")
                self._lifecycle = UsercodeLifecycle(
                    uuid,
                    disk_info,
                    self.update_status,
                    self._log_helper,
                    self.config,
                )
                asyncio.ensure_future(self._lifecycle.run_process())
            else:
                LOGGER.warn("Cannot run usercode, there is already a lifecycle present.")
                with disk_info.mount_path.joinpath("log.txt").open("w") as fh:
                    fh.write("Unable to start code.\n")
                    fh.write("It is not safe to run multiple code disks at once.\n")

    async def handle_disk_removal(self, uuid: DiskUUID, disk_info: DiskInfo) -> None:
        """Handle a disk removal."""
        LOGGER.debug(f"Disk removed: {uuid} ({disk_info.disk_type})")

        if disk_info.disk_type is DiskType.USERCODE:
            LOGGER.info(f"Usercode disk {uuid} removed ({disk_info.mount_path})")

            if self._lifecycle is not None and self._lifecycle._uuid == disk_info.uuid:
                await self._lifecycle.kill_process()
                self._lifecycle = None
                self.update_status()
            else:
                LOGGER.warning("Disk removed, but no code lifecycle available")

    async def handle_kill_request(
            self,
            request: UsercodeKillManagerRequest,
    ) -> RequestResponse:
        """Handle a request to kill running usercode."""
        if self._lifecycle is None:
            return RequestResponse(
                uuid=request.uuid,
                success=False,
                reason="No active usercode lifecycle",
            )
        else:
            LOGGER.info("Kill request received.")
            await self._lifecycle.kill_process()
            return RequestResponse(
                uuid=request.uuid,
                success=True,
            )

    async def handle_restart_request(
        self,
        request: UsercodeRestartManagerRequest,
    ) -> RequestResponse:
        """Handle a request to restart usercode."""
        LOGGER.info("Restart request received.")
        if self._lifecycle is None:
            return RequestResponse(
                uuid=request.uuid,
                success=False,
                reason="No active usercode lifecycle",
            )
        else:
            if self._lifecycle.status is CodeStatus.RUNNING:
                return RequestResponse(
                    uuid=request.uuid,
                    success=False,
                    reason="Code is already running.",
                )
            else:
                asyncio.ensure_future(self._lifecycle.run_process())
                return RequestResponse(
                    uuid=request.uuid,
                    success=True,
                )

    def update_status(self, code_status: Optional[CodeStatus] = None) -> None:
        """
        Calculate and update the status of this manager.

        Called by the usercode lifecycle to inform us of changes.
        """
        if self._lifecycle is None:
            # When the status is updated in the lifecycle constructor, we
            # are left with a situation where there is no lifecycle object,
            # but the code is starting. Thus we want to inform anyway.
            #
            # This section also updates the status when the lifecycle is cleaned up.
            self.status = ProcessManagerMessage(
                status=ProcessManagerMessage.Status.RUNNING,
                code_status=code_status,
            )
        else:
            self.status = ProcessManagerMessage(
                status=ProcessManagerMessage.Status.RUNNING,
                code_status=self._lifecycle.status,
                disk_info=self._lifecycle.disk_info,
            )



"""MQTT Wrapper."""

import asyncio
import logging
from json import JSONDecodeError, loads
from typing import (
    Any,
    Callable,
    Coroutine,
    Dict,
    List,
    Match,
    Optional,
    TypeVar,
)
from uuid import UUID

import gmqtt
from pydantic import BaseModel

from astoria.common.config import MQTTBrokerInfo
from astoria.common.manager_requests import ManagerRequest, RequestResponse
from astoria.common.messages.base import ManagerMessage

from .topic import Topic

LOGGER = logging.getLogger(__name__)

Handler = Callable[[Match[str], str], Coroutine[Any, Any, None]]  # type: ignore
RequestT = TypeVar("RequestT", bound=ManagerRequest)


class MQTTWrapper:
    """
    MQTT wrapper class.

    Wraps the functionality that we are using for MQTT, with extra
    sanity checks and validation to make sure that things are less
    likely to go wrong.
    """

    _client: gmqtt.Client

    def __init__(
        self,
        client_name: str,
        broker_info: MQTTBrokerInfo,
        *,
        last_will: Optional[BaseModel] = None,
        dependencies: List[str] = [],
        no_dependency_event: Optional[asyncio.Event] = None,
    ) -> None:
        self._client_name = client_name
        self._broker_info = broker_info
        self._last_will = last_will
        self._dependencies = dependencies
        self._no_dependency_event = no_dependency_event

        self._dependency_events: Dict[str, asyncio.Event] = {
            name: asyncio.Event() for name in self._dependencies
        }

        self._topic_handlers: Dict[Topic, Handler] = {}

        self._client = gmqtt.Client(
            self._client_name,
            will_message=self.last_will_message,
        )

        self._client.reconnect_retries = 0

        self._client.on_message = self.on_message
        self._client.on_connect = self.on_connect
        self._client.on_disconnect = self.on_disconnect

        self.subscribe("+", self._dependency_message_handler)

        # Subscribe to request response from dependent managers
        for manager in self._dependencies:
            self.subscribe(
                f"{manager}/request/+/+",
                self._request_response_message_handler,
            )
        self._request_response_events: Dict[UUID, asyncio.Event] = {}
        self._request_response_data: Dict[UUID, RequestResponse] = {}

    @property
    def is_connected(self) -> bool:
        """Determine if the client connected to the broker."""
        return self._client.is_connected

    @property
    def last_will_message(self) -> Optional[gmqtt.Message]:
        """Last will and testament message for this client."""
        if self._last_will is not None:
            return gmqtt.Message(
                self.mqtt_prefix,
                self._last_will.json(),
                retain=True,
            )
        else:
            return None

    @property
    def mqtt_prefix(self) -> str:
        """The topic prefix for MQTT."""
        return f"{self._broker_info.topic_prefix}/{self._client_name}"

    async def connect(self) -> None:
        """Connect to the broker."""
        if self.is_connected:
            LOGGER.error("Attempting connection, but client is already connected.")
        mqtt_version = gmqtt.constants.MQTTv50
        if self._broker_info.force_protocol_version_3_1:
            mqtt_version = gmqtt.constants.MQTTv311

        await self._client.connect(
            self._broker_info.host,
            port=self._broker_info.port,
            ssl=self._broker_info.enable_tls,
            version=mqtt_version,
        )

    async def disconnect(self) -> None:
        """Disconnect from the broker."""
        if not self.is_connected:
            LOGGER.error(
                "Attempting disconnection, but client is already disconnected.",
            )

        await self._client.disconnect()

        if self.is_connected:
            raise RuntimeError("Disconnection was attempted, but was unsuccessful")

    def on_connect(
        self,
        client: gmqtt.client.Client,
        flags: int,
        rc: int,
        properties: Dict[str, List[int]],
    ) -> None:
        """Callback for mqtt connection."""
        for topic in self._topic_handlers:
            LOGGER.debug(f"Subscribing to {topic}")
            client.subscribe(str(topic))

    def on_disconnect(self, client: gmqtt.client.Client, packet: bytes) -> None:
        """Callback for mqtt disconnection."""
        LOGGER.debug("MQTT client disconnected")
        if self._no_dependency_event is not None:
            self._no_dependency_event.set()

    async def on_message(
        self,
        client: gmqtt.client.Client,
        topic: str,
        payload: bytes,
        qos: int,
        properties: Dict[str, int],
    ) -> gmqtt.constants.PubRecReasonCode:
        """Callback for mqtt messages."""
        LOGGER.debug(f"Message received on {topic} with payload: {payload!r}")
        for t, handler in self._topic_handlers.items():
            match = t.match(topic)
            if match:
                asyncio.ensure_future(self.wait_dependencies())
                LOGGER.debug(f"Calling {handler.__name__} to handle {topic}")
                asyncio.ensure_future(handler(match, payload.decode()))

        return gmqtt.constants.PubRecReasonCode.SUCCESS

    def publish(
        self,
        topic: str,
        payload: BaseModel,
        *,
        retain: bool = False,
        auto_prefix_topic: bool = True,
        auto_prefix_client_name: bool = True,
    ) -> None:
        """Publish a payload to the broker."""
        if not self.is_connected:
            LOGGER.error(
                "Attempted to publish message, but client is not connected.",
            )

        if auto_prefix_client_name:
            prefix = self.mqtt_prefix
        else:
            prefix = self._broker_info.topic_prefix

        if len(topic) == 0:
            topic_complete = Topic.parse(prefix)
        elif auto_prefix_topic:
            topic_complete = Topic.parse(f"{prefix}/{topic}")
        else:
            topic_complete = Topic.parse(topic)

        if not topic_complete.is_publishable:
            raise ValueError(f"Cannot publish to MQTT topic: {topic_complete}")

        self._client.publish(
            str(topic_complete),
            payload.json(),
            qos=1,
            retain=retain,
        )

    def subscribe(
        self,
        topic: str,
        callback: Handler,
    ) -> None:
        """
        Subscribe to an MQTT Topic.

        Callback is called when a message arrives.

        Should be called before the MQTT wrapper is connected.
        """
        if len(topic) == 0:
            topic_complete = Topic.parse(self.mqtt_prefix)
        else:
            topic_complete = Topic.parse(f"{self._broker_info.topic_prefix}/{topic}")

        self._topic_handlers[topic_complete] = callback

    async def wait_dependencies(self) -> None:
        """Wait for all dependencies."""
        if len(self._dependencies) > 0:
            LOGGER.debug("Waiting for " + ", ".join(self._dependencies))

            tasks = [asyncio.gather(
                *(event.wait() for event in self._dependency_events.values()),
            )]

            if self._no_dependency_event is not None:
                tasks.append(self._no_dependency_event.wait())  # type: ignore

            await asyncio.wait(
                tasks,
                return_when=asyncio.FIRST_COMPLETED,
            )

    @property
    def has_dependencies(self) -> bool:
        """Are the dependencies of the manager available?."""
        return all(event.is_set() for event in self._dependency_events.values())

    async def _dependency_message_handler(
        self,
        match: Match[str],
        payload: str,
    ) -> None:
        """Handle status messages from state managers."""
        manager = match.group(1)
        try:
            info = ManagerMessage(**loads(payload))
            LOGGER.debug(f"Status update from {manager}: {info.status}")
            if info.status is ManagerMessage.Status.RUNNING:
                try:
                    self._dependency_events[manager].set()
                except KeyError:
                    pass
            else:
                if self.has_dependencies and manager in self._dependency_events:
                    LOGGER.warning(f"{manager} is unavailable!")
                    if self._no_dependency_event is not None:
                        self._no_dependency_event.set()
        except JSONDecodeError:
            pass

    async def manager_request(
        self,
        manager: str,
        request_name: str,
        request: RequestT,
        *,
        response_timeout: float = 1,
    ) -> RequestResponse:
        """
        Perform a manager request.

        Raises exception if not dependent on manager, or if request fails.
        """
        if manager not in self._dependencies:
            raise ValueError(
                f"{manager} must be listed as dependency to make manager request",
            )

        topic = f"{self._broker_info.topic_prefix}/{manager}/request/{request_name}"

        self._request_response_events[request.uuid] = asyncio.Event()

        self.publish(topic, request, auto_prefix_topic=False)

        try:
            await asyncio.wait_for(
                self._request_response_events[request.uuid].wait(),
                2,
            )
        except asyncio.TimeoutError as e:
            raise RuntimeError("No response to manager request") from e
        if request.uuid not in self._request_response_data:
            raise RuntimeError("Request Response not available.")
        else:
            self._request_response_events.pop(request.uuid)
            return self._request_response_data.pop(request.uuid)

    async def _request_response_message_handler(
        self,
        match: Match[str],
        payload: str,
    ) -> None:
        """Handle request response messages."""
        uuid = UUID(match.group(2))
        # If uuid not recognised, probably a response for another client
        if uuid in self._request_response_events:
            try:
                self._request_response_data[uuid] = RequestResponse(
                    **loads(payload),
                )
            except JSONDecodeError:
                self._request_response_data[uuid] = RequestResponse(
                    uuid=uuid,
                    success=False,
                    reason="Unable to decode request response.",
                )
            self._request_response_events[uuid].set()



"""Commands for managing static disks."""
import click

from .add import add
from .remove import remove
from .remove_all import remove_all


@click.group("static-disk")
def static_disk() -> None:
    """Interact with metadata."""


static_disk.add_command(add)
static_disk.add_command(remove)
static_disk.add_command(remove_all)



"""
Configuration schema for Astoria.

Common to all components.
"""
from pathlib import Path
from typing import IO, List, Optional

from pydantic import BaseModel
from toml import load


class MQTTBrokerInfo(BaseModel):
    """MQTT Broker Information."""

    host: str
    port: int
    enable_tls: bool = False
    topic_prefix: str = "astoria"
    force_protocol_version_3_1: bool = False

    class Config:
        """Pydantic config."""

        extra = "forbid"


class WiFiInfo(BaseModel):
    """System settings for WiFi."""

    interface: str
    bridge: str
    enable_wpa3: bool

    class Config:
        """Pydantic config."""

        extra = "forbid"


class SystemInfo(BaseModel):
    """System settings that don't find elsewhere."""

    cache_dir: Path
    initial_log_lines: List[str] = []

    class Config:
        """Pydantic config."""

        extra = "forbid"


class AstoriaConfig(BaseModel):
    """Config schema for Astoria."""

    mqtt: MQTTBrokerInfo
    wifi: WiFiInfo
    system: SystemInfo

    class Config:
        """Pydantic config."""

        extra = "forbid"

    @classmethod
    def _get_config_path(cls, config_str: Optional[str] = None) -> Path:
        """Check for a config file or search the filesystem for one."""
        CONFIG_SEARCH_PATHS = [
            Path("astoria.toml"),
            Path("/etc/astoria.toml"),
        ]
        if config_str is None:
            for path in CONFIG_SEARCH_PATHS:
                if path.exists() and path.is_file():
                    return path
        else:
            path = Path(config_str)
            if path.exists() and path.is_file():
                return path
        raise FileNotFoundError("Unable to find config file.")

    @classmethod
    def load(cls, config_str: Optional[str] = None) -> 'AstoriaConfig':
        """Load the config."""
        config_path = cls._get_config_path(config_str)
        with config_path.open("r") as fh:
            return cls.load_from_file(fh)

    @classmethod
    def load_from_file(cls, fh: IO[str]) -> 'AstoriaConfig':
        """Load the config from a file."""
        return cls(**load(fh))



"""Types for astmetad."""
import platform
import random
import re
import secrets
from enum import Enum
from typing import Optional

from pydantic import BaseModel, validator

from astoria import __version__
from astoria.common.config import AstoriaConfig

from .base import ManagerMessage

SSID_PREFIX = "robot-"
MAX_SSID_LENGTH = 32  # SSIDs must be no more than 32 octets.


class RobotMode(Enum):
    """Running Status of the manager daemon."""

    COMP = "COMP"
    DEV = "DEV"


class Metadata(BaseModel):
    """Astoria Metadata."""

    class Config:
        """Pydantic config."""

        validate_assignment = True

    @classmethod
    def init(cls, config: AstoriaConfig) -> 'Metadata':
        """
        Initialise the metadata and populate with data.

        Based on information from software and the astoria configuration.
        """
        uname = platform.uname()
        return cls(
            astoria_version=__version__,
            kernel_version=uname.release,
            arch=uname.machine,
            python_version=platform.python_version(),
            libc_ver="".join(platform.libc_ver()),
        )

    def is_wifi_valid(self) -> bool:
        """
        Check if the WiFi configuration is valid to be turned on.

        Checks that the WiFi is enabled and that the config params are set.Optional
        :return: boolean indicating whether the WiFi can be turned on.
        """
        return all([
            self.wifi_enabled,
            self.wifi_ssid is not None,
            self.wifi_psk is not None,
            self.wifi_region is not None,
        ])

    # From Meta USB
    arena: str = "A"
    zone: int = 0
    mode: RobotMode = RobotMode.DEV
    marker_offset: int = 0
    game_timeout: Optional[int] = None
    wifi_enabled: bool = True

    # From Software
    astoria_version: str
    kernel_version: str
    arch: str
    python_version: str
    libc_ver: str

    # From astoria.toml
    wifi_ssid: Optional[str] = None
    wifi_psk: Optional[str] = None
    wifi_region: Optional[str] = None


class RobotSettings(BaseModel):
    """Schema for robot-settings.toml."""

    team_tla: str
    wifi_psk: str
    wifi_region: str = "GB"  # Assume GB as that is where most competitors are.
    wifi_enabled: bool = True

    @validator("team_tla")
    def validate_team_tla(cls, val: str) -> str:
        """
        Validate the TLA.

        :param val: The received TLA value.
        :returns: The TLA value.
        """
        if val == "beeeeees":
            return "🐝" * 6

        if not re.match(r"^[A-Z]{3}\d*$", val, re.IGNORECASE):
            raise ValueError("Team name did not match format.")

        if len(val) > MAX_SSID_LENGTH - len(SSID_PREFIX):
            raise ValueError(
                f"SSID {SSID_PREFIX}{val} is longer than "
                f"maximum length: {MAX_SSID_LENGTH} octets.",
            )

        return val.upper()

    @classmethod
    def generate_default_settings(cls) -> 'RobotSettings':
        """Generate default sensible settings for the robot."""
        random_tla = f"ZZZ{random.randint(0, 99999)}"

        # Use random characters for the WiFi password as passphrase schemes
        # such as Diceware are very language specific. This can be changed
        # by competitors if they need to.
        #
        # Also separate every 4 characters by a dash for readability
        passphrase = "-".join(secrets.token_hex(2) for _ in range(3))

        return cls(
            team_tla=random_tla,
            wifi_psk=passphrase,
        )


class MetadataManagerMessage(ManagerMessage):
    """
    Status message for Metadata Manager.

    Published to /astoria/astmetad
    """

    metadata: Metadata



"""Helper class to manage broadcast events."""
import logging
from asyncio import PriorityQueue
from json import JSONDecodeError, loads
from typing import TYPE_CHECKING, Any, Generic, Match, Type, TypeVar

from astoria.common.broadcast_event import BroadcastEvent

if TYPE_CHECKING:
    from .wrapper import MQTTWrapper

LOGGER = logging.getLogger(__name__)

T = TypeVar("T", bound=BroadcastEvent)


class BroadcastHelper(Generic[T]):
    """Helper class to manager broadcast events."""

    def __init__(self, mqtt: 'MQTTWrapper', name: str, schema: Type[T]) -> None:
        self._mqtt = mqtt
        self._name = name
        self._schema = schema

        self._event_queue: PriorityQueue[T] = PriorityQueue()
        self._mqtt.subscribe(f"broadcast/{name}", self._handle_broadcast)

    @classmethod
    def get_helper(cls, mqtt: 'MQTTWrapper', schema: Type[T]) -> 'BroadcastHelper[T]':
        """Get the broadcast helper for a given event."""
        return BroadcastHelper[T](mqtt, schema.name, schema)

    async def _handle_broadcast(
        self,
        match: Match[str],
        payload: str,
    ) -> None:
        """
        Handle a broadcast event message.

        Inserts the event inserts it into the priority queue.
        """
        try:
            ev = self._schema(**loads(payload))
            LOGGER.debug(
                f"Received {ev.event_name} broadcast event from {ev.sender_name}",
            )
            await self._event_queue.put(ev)
        except JSONDecodeError:
            LOGGER.warning(f"Broadcast event {self._name} contained invalid JSON")

    def send(self, **kwargs: Any) -> None:  # type: ignore
        """Send an event."""
        data = self._schema(
            event_name=self._schema.name,
            sender_name=self._mqtt._client_name,
            **kwargs,
        )
        self._mqtt.publish(
            f"broadcast/{self._schema.name}",
            data,
            auto_prefix_client_name=False,
        )

    async def wait_broadcast(self) -> T:
        """
        Wait for an event on the given broadcast.

        :param broadcast_name: The name of the broadcast to wait for.
        """
        return await self._event_queue.get()



"""
Astoria command line interface.

Split up into one class per command.
"""
import click

from .list_disks import list_disks
from .metadata import metadata
from .static_disks import static_disk
from .usercode import usercode


@click.group("astdiskd")
def main() -> None:
    """Astoria Command Line Utility."""


main.add_command(list_disks)
main.add_command(metadata)
main.add_command(usercode)
main.add_command(static_disk)

if __name__ == "__main__":
    main()



"""Command to restart running usercode."""
import asyncio
from typing import Optional

import click

from astoria.common.manager_requests import ManagerRequest
from astoria.consumers.astctl.command import Command

loop = asyncio.get_event_loop()


@click.command("restart")
@click.option("-v", "--verbose", is_flag=True)
@click.option("-c", "--config-file", type=click.Path(exists=True))
def restart(*, verbose: bool, config_file: Optional[str]) -> None:
    """Restart running usercode."""
    command = RestartUsercodeCommand(verbose, config_file)
    loop.run_until_complete(command.run())


class RestartUsercodeCommand(Command):
    """Restart running usercode."""

    dependencies = ["astprocd"]

    async def main(self) -> None:
        """Main method of the command."""
        res = await self._mqtt.manager_request(
            "astprocd",
            "restart",
            ManagerRequest(sender_name=self.name),
        )
        if res.success:
            print("Successfully restarted code.")
            if len(res.reason) > 0:
                print(res.reason)
        else:
            print("Unable to restart code.")
            if len(res.reason) > 0:
                print(res.reason)
        # Add timeout
        self.halt(silent=True)



"""Commands to interact with metadata."""
import click

from .set import set
from .show import show


@click.group("metadata")
def metadata() -> None:
    """Interact with metadata."""


metadata.add_command(set)
metadata.add_command(show)



"""Commands to interact with usercode."""
import click

from .kill import kill
from .log import log
from .restart import restart
from .show import show
from .trigger import trigger


@click.group("usercode")
def usercode() -> None:
    """Interact with Usercode."""


usercode.add_command(kill)
usercode.add_command(log)
usercode.add_command(restart)
usercode.add_command(show)
usercode.add_command(trigger)



"""Cache metadata attributes on disk."""

import logging
from json import JSONDecodeError, dumps, loads
from pathlib import Path
from typing import Dict, Optional, Set

LOGGER = logging.getLogger(__name__)


class MetadataCache:
    """Manages loading and updating data from the metadata cache."""

    def __init__(
        self,
        cached_keys: Set[str],
        *,
        cache_path: Path = Path("/var/srobo/astmetad-cache.json"),
    ) -> None:
        """
        Construct the metadata cache.

        :param cached_keys: Keys to allow in the cache.
        :param cache_path: Path to the cache file in an appropriate location.
        """
        self._cached_keys = cached_keys
        self._cache_path = cache_path
        self._ensure_cache_exists()
        self._data = self._read_cache()

    def _ensure_cache_exists(self) -> None:
        """
        Check whether the metadata cache exists.

        If the cache does not exist, it will be created with no data.
        """
        LOGGER.info(f"Cache file is {self._cache_path}")
        if not self._cache_path.exists():
            LOGGER.debug("Cache file does not exist!")

            # Ensure parent directory exists
            self._cache_path.parent.mkdir(parents=True, exist_ok=True)

            LOGGER.info("Creating new cache file.")
            self._write_cache({})

    def _read_cache(self) -> Dict[str, str]:
        """
        Read the data from the cache file.

        Ignore any data that isn't permitted.

        :return: Data from the cache file.
        """
        LOGGER.debug("Reading data from cache.")
        with self._cache_path.open("r") as fh:
            raw_data = fh.read()

        try:
            unvalidated_data = loads(raw_data)

            # Validate that everything is a string.
            for k, v in unvalidated_data.items():
                if not isinstance(k, str) or not isinstance(v, str):
                    LOGGER.warning(f"Invalid JSON data in cache: {raw_data}")
                    return {}

            data = {k: v for k, v in unvalidated_data.items() if k in self._cached_keys}

            if data != unvalidated_data:
                LOGGER.warning("Bad keys in cache, re-writing cache!")
                self._write_cache(data)
            return data

        except JSONDecodeError:
            LOGGER.warning(f"Invalid JSON data in cache: {raw_data}")
            return {}

    def _write_cache(self, data: Dict[str, str]) -> None:
        """
        Write the data to the cache file.

        :param data: Data to write to the cache file.
        """
        LOGGER.debug(f"Writing data to cache: {data}")
        with self._cache_path.open("w") as fh:
            fh.write(dumps(data))

    @property
    def data(self) -> Dict[str, str]:
        """The cached data."""
        return self._data

    def update_cached_attr(self, key: str, value: Optional[str]) -> None:
        """
        Update a cached attribute.

        :param key: The key of the attribute to update.
        :param value: The value to give the attribute.

        :raises ValueError: Key is not permitted in cache.
        """
        if key not in self._cached_keys:
            raise ValueError(
                f"Tried to cache {key}, but it is not allowed to be cached.",
            )
        else:
            if key not in self._data or self._data[key] != value:
                if value is None:
                    LOGGER.info(f"Deleting {key} from cache.")
                    self._data.pop(key, None)  # Remove the key from the cache.
                else:
                    LOGGER.info(f"Updated cache: {key} -> {value}")
                    self._data[key] = str(value)  # Make sure the value is a string
                self._write_cache(self._data)



"""Message schemas for astprocd."""
from enum import Enum
from typing import Optional

from .astdiskd import DiskInfo
from .base import ManagerMessage


class CodeStatus(str, Enum):
    """Status of the running code."""

    STARTING = "code_starting"
    RUNNING = "code_running"
    KILLED = "code_killed"
    FINISHED = "code_finished"
    CRASHED = "code_crashed"


class ProcessManagerMessage(ManagerMessage):
    """
    Status message for Process Manager.

    Published to astoria/astprocd
    """

    code_status: Optional[CodeStatus]
    disk_info: Optional[DiskInfo]



"""Command to add a filesystem path as a static disk."""
import asyncio
from pathlib import Path
from typing import Optional

import click

from astoria.common.manager_requests import RemoveStaticDiskRequest
from astoria.consumers.astctl.command import Command

loop = asyncio.get_event_loop()


@click.command("remove")
@click.argument("path")
@click.option("-v", "--verbose", is_flag=True)
@click.option("-c", "--config-file", type=click.Path(exists=True))
def remove(path: str, *, verbose: bool, config_file: Optional[str]) -> None:
    """Unmount a static disk."""
    command = RemoveStaticDiskCommand(path, verbose, config_file)
    loop.run_until_complete(command.run())


class RemoveStaticDiskCommand(Command):
    """Command to add a filesystem path as a static disk."""

    _path: Path

    dependencies = ["astdiskd"]

    def __init__(
        self,
        path: str,
        verbose: bool,
        config_file: Optional[str],
    ):
        super().__init__(verbose, config_file)
        self._path = Path(path).resolve()

    async def main(self) -> None:
        """Main method of the command."""
        res = await self._mqtt.manager_request(
            "astdiskd",
            "remove_static_disk",
            RemoveStaticDiskRequest(sender_name=self.name, path=self._path),
        )
        if res.success:
            print("Successfully removed disk.")
            if len(res.reason) > 0:
                print(res.reason)
        else:
            print("Unable to remove disk.")
            if len(res.reason) > 0:
                print(res.reason)
        # Add timeout
        self.halt(silent=True)



"""
MQTT Topic Abstraction.

Allows topic strings to be constructed, along with regex to match them.
"""

from re import compile
from typing import Dict, Match, Optional, Pattern, Sequence


class Topic:
    """
    An MQTT Topic.

    A topic that may be published or subscribed to.
    """

    WILDCARDS: Dict[str, str] = {
        "+": "([^/]+)",
        "#": "(.+)",
    }

    def __init__(self, parts: Sequence[str]) -> None:
        self.parts = parts

    def match(self, topic: str) -> Optional[Match[str]]:
        """Perform a regex match on a topic."""
        return self.regex.match(topic)

    @classmethod
    def parse(cls, topic: str) -> 'Topic':
        """
        Parse a string topic into a Topic instance.

        It is assumed that the supplied topic complies with the MQTT spec.
        If not, then behaviour is undefined.
        """
        if len(topic) > 1:
            if topic[:1] == "/" or topic[-1:] == "/":
                raise ValueError("Topic cannot begin or end with /")
        else:
            if topic == "/" or len(topic) == 0:
                raise ValueError("Invalid Topic")

        return cls(topic.split("/"))

    def __str__(self) -> str:
        return "/".join(str(p) for p in self.parts)

    def __repr__(self) -> str:
        return f"Topic(\"{self}\")"

    def __hash__(self) -> int:
        return hash(repr(self))

    def __eq__(self, other: object) -> bool:
        try:
            return self.parts == other.parts  # type: ignore
        except AttributeError:
            return False

    @property
    def is_publishable(self) -> bool:
        """
        Determine if it is valid to publish to this topic.

        We are not permitting publication to topics containing wildcards.
        """
        return all(x not in self.parts for x in self.WILDCARDS)

    @property
    def regex(self) -> Pattern[str]:
        """
        Regular expression to match the topic.

        Any wildcard fields are available as capture groups.
        """
        handled_parts = []
        for p in self.parts:
            try:
                handled_parts.append(self.WILDCARDS[p])
            except KeyError:
                handled_parts.append(p)

        return compile("^" + "/".join(handled_parts) + "$")



"""Command to virtually press start button."""
import asyncio
from typing import Optional

import click

from astoria.common.broadcast_event import StartButtonBroadcastEvent
from astoria.common.mqtt import BroadcastHelper
from astoria.consumers.astctl.command import Command

loop = asyncio.get_event_loop()


@click.command("trigger")
@click.option("-v", "--verbose", is_flag=True)
@click.option("-c", "--config-file", type=click.Path(exists=True))
def trigger(*, verbose: bool, config_file: Optional[str]) -> None:
    """Virtually trigger the start button."""
    command = TriggerUsercodeCommand(verbose, config_file)
    loop.run_until_complete(command.run())


class TriggerUsercodeCommand(Command):
    """Virtually trigger the start button."""

    def _init(self) -> None:
        """
        Initialisation of the data component.

        Called in the constructor of the parent class.
        """
        self._trigger_event = BroadcastHelper.get_helper(
            self._mqtt,
            StartButtonBroadcastEvent,
        )

    async def main(self) -> None:
        """Send a trigger event."""
        self._trigger_event.send()



"""Process Manager Application."""

import asyncio
import logging
from typing import Optional

import click

from .process_manager import ProcessManager

LOGGER = logging.getLogger(__name__)

loop = asyncio.get_event_loop()


@click.command("astprocd")
@click.option("-v", "--verbose", is_flag=True)
@click.option("-c", "--config-file", type=click.Path(exists=True))
def main(*, verbose: bool, config_file: Optional[str]) -> None:
    """Process Manager Application Entrypoint."""
    testd = ProcessManager(verbose, config_file)
    loop.run_until_complete(testd.run())


if __name__ == "__main__":
    main()



"""State Manager base class."""
import asyncio
import logging
from abc import ABCMeta, abstractmethod
from json import JSONDecodeError, loads
from typing import Callable, Coroutine, Generic, Match, Type, TypeVar

from .data_component import DataComponent
from .manager_requests import ManagerRequest, RequestResponse
from .messages.base import ManagerMessage

LOGGER = logging.getLogger(__name__)

loop = asyncio.get_event_loop()

T = TypeVar("T", bound=ManagerMessage)
RequestT = TypeVar("RequestT", bound=ManagerRequest)


class StateManager(DataComponent, Generic[T], metaclass=ABCMeta):
    """
    State Manager.

    A process that stores and mutates some state.
    """

    _status: T

    @property
    def status(self) -> T:
        """Get the status of the state manager."""
        return self._status

    @status.setter
    def status(self, status: T) -> None:
        """Set the status of the state manager."""
        self._status = status
        self._mqtt.publish("", status, retain=True)

    @property
    @abstractmethod
    def offline_status(self) -> T:
        """
        Status to publish when the manager goes offline.

        This status should ensure that any other components relying
        on this data go into a safe state.
        """
        raise NotImplementedError

    @property
    def last_will(self) -> T:
        """Last will and testament of the MQTT client."""
        return self.offline_status

    async def _post_connect(self) -> None:
        """Overridable callback after MQTT connection."""
        LOGGER.info("Ready.")
        await self._mqtt.wait_dependencies()

    async def _pre_disconnect(self) -> None:
        """Change status to offline before disconnecting from the broker."""
        self.status = self.offline_status

    def _register_request(
        self,
        name: str,
        typ: Type[RequestT],
        handler: Callable[[RequestT], Coroutine[None, None, RequestResponse]],
    ) -> None:
        LOGGER.debug(f"Registering {name} request for {self.name} component")

        async def _handler(match: Match[str], payload: str) -> None:
            try:
                req = typ(**loads(payload))
                response = await handler(req)
                self._mqtt.publish(
                    f"request/{name}/{req.uuid}",
                    response,
                )
            except JSONDecodeError:
                LOGGER.warning(f"Received {name} request, but unable to decode JSON")

        self._mqtt.subscribe(f"{self.name}/request/{name}", _handler)



"""Lifecycle classes to load metadata from disks."""

import asyncio
import logging
from abc import ABCMeta, abstractmethod
from json import JSONDecodeError, loads
from pathlib import Path
from typing import Dict

import toml
from pydantic.error_wrappers import ValidationError

from astoria.common.messages.astdiskd import DiskInfo, DiskUUID
from astoria.common.messages.astmetad import SSID_PREFIX, RobotSettings

LOGGER = logging.getLogger(__name__)

loop = asyncio.get_event_loop()


class AbstractMetadataDiskLifecycle(metaclass=ABCMeta):
    """Load and validate metadata from a disk."""

    def __init__(self, uuid: DiskUUID, disk_info: DiskInfo) -> None:
        self._uuid = uuid
        self._disk_info = disk_info

        self._diff = self.extract_diff_data()

    @abstractmethod
    def extract_diff_data(self) -> Dict[str, str]:
        """Extract the diff data from the disk."""
        raise NotImplementedError

    @property
    def diff_data(self) -> Dict[str, str]:
        """The data to be used as override."""
        return self._diff


class MetadataDiskLifecycle(AbstractMetadataDiskLifecycle):
    """Load and validate metadata from a JSON file on the disk."""

    def extract_diff_data(self) -> Dict[str, str]:
        """
        Extract the diff data from the disk.

        Loads astoria.json from the disk and parses it as JSON.
        """
        metadata_file_path = self._disk_info.mount_path / "astoria.json"

        try:
            with metadata_file_path.open("r") as fh:
                return loads(fh.read())  # type: ignore
        except FileNotFoundError:
            LOGGER.warning("Unable to find metadata.json.")
        except JSONDecodeError:
            LOGGER.warning("Invalid JSON in astoria.json")
        return {}


class NoValidRobotSettingsException(Exception):
    """The robot settings were not valid or did not exist."""


class UsercodeDiskLifecycle(AbstractMetadataDiskLifecycle):
    """Load and validate metadata from a usercode disk."""

    def _load_settings_file(self, path: Path) -> RobotSettings:
        """
        Load the robot settings file.

        :param path: The file to load settings from.
        :raises NoValidRobotSettingsException: The robot settings were not valid.
        :returns: The robot settings in path.
        """
        if not path.exists():
            raise NoValidRobotSettingsException("File does not exist.")

        try:
            data = toml.load(path)
        except toml.TomlDecodeError:
            raise NoValidRobotSettingsException("Invalid TOML")

        try:
            return RobotSettings(**data)
        except ValidationError as e:
            raise NoValidRobotSettingsException(
                f"Settings did not match schema: {e}",
            )

    def extract_diff_data(self) -> Dict[str, str]:
        """
        Extract the diff data from the disk.

        Loads robot-settings.toml fom the disk.
        """
        robot_settings_file = self._disk_info.mount_path / "robot-settings.toml"
        try:
            settings = self._load_settings_file(robot_settings_file)
        except NoValidRobotSettingsException:
            settings = RobotSettings.generate_default_settings()

            LOGGER.warning("No valid settings, writing sensible defaults.")
            with robot_settings_file.open("w") as fh:
                toml.dump(settings.dict(), fh)

        return {
            "wifi_ssid": SSID_PREFIX + settings.team_tla,
            "wifi_psk": settings.wifi_psk,
            "wifi_region": settings.wifi_region,
            "wifi_enabled": str(settings.wifi_enabled),
        }



"""A provider of disk information."""

from abc import ABCMeta
from pathlib import Path
from typing import TYPE_CHECKING, Callable, Coroutine, Dict

from astoria.common.messages.astdiskd import DiskUUID

if TYPE_CHECKING:
    from .disk_manager import DiskManager


class DiskProvider(metaclass=ABCMeta):
    """
    A provider of disk information.

    Multiple disk providers can be used within one astdiskd process.
    """

    def __init__(
        self,
        disk_manager: 'DiskManager',
        *,
        notify_coro: Callable[[], Coroutine[None, None, None]],
    ) -> None:
        self._disk_manager = disk_manager
        self._notify_coro = notify_coro

        self._disks: Dict[DiskUUID, Path] = {}

    @property
    def disks(self) -> Dict[DiskUUID, Path]:
        """Currently mounted disks."""
        return self._disks

    async def main(self) -> None:
        """Main loop to detect disks if necessary."""
        pass



"""Command to show metadata."""
import asyncio
from typing import Optional

import click

from astoria.common.messages.astmetad import MetadataManagerMessage
from astoria.consumers.astctl.command import SingleManagerMessageCommand

loop = asyncio.get_event_loop()


@click.command("show")
@click.option("-v", "--verbose", is_flag=True)
@click.option("-c", "--config-file", type=click.Path(exists=True))
def show(*, verbose: bool, config_file: Optional[str]) -> None:
    """Show current metadata."""
    command = ShowMetadataCommand(verbose, config_file)
    loop.run_until_complete(command.run())


class ShowMetadataCommand(SingleManagerMessageCommand[MetadataManagerMessage]):
    """Show current metadata."""

    manager = "astmetad"
    message_schema = MetadataManagerMessage

    def handle_message(
        self,
        message: MetadataManagerMessage,
    ) -> None:
        """Print the metadata."""
        print("Current Astoria Metadata is:")
        for i, v in sorted(message.metadata.__dict__.items()):
            print(f"\t{i}: {v}")



"""Common data for MQTT messages."""
from enum import Enum

from pydantic import BaseModel

from astoria import __version__


class ManagerMessage(BaseModel):
    """Common data that all manager messages output."""

    class Status(Enum):
        """Running Status of the manager daemon."""

        STOPPED = "STOPPED"
        RUNNING = "RUNNING"

    status: Status
    astoria_version: str = __version__



"""Command to view usercode logs in real-time."""
import asyncio
from typing import Optional

import click

from astoria.common.broadcast_event import UsercodeLogBroadcastEvent
from astoria.common.mqtt import BroadcastHelper
from astoria.consumers.astctl.command import Command

loop = asyncio.get_event_loop()


@click.command("log")
@click.option("-v", "--verbose", is_flag=True)
@click.option("-c", "--config-file", type=click.Path(exists=True))
def log(*, verbose: bool, config_file: Optional[str]) -> None:
    """View usercode logs in real-time."""
    command = ViewUsercodeLogCommand(verbose, config_file)
    loop.run_until_complete(command.run())


class ViewUsercodeLogCommand(Command):
    """Command to view usercode logs in real-time."""

    def _init(self) -> None:
        """
        Initialisation of the data component.

        Called in the constructor of the parent class.
        """
        self._log_event = BroadcastHelper.get_helper(
            self._mqtt,
            UsercodeLogBroadcastEvent,
        )

    async def main(self) -> None:
        """Send a trigger event."""
        while not self._stop_event.is_set():
            # wait_broadcast waits forever until a broadcoast, so we will use a short
            # timeout to ensure that the loop condition is checked.
            try:
                ev = await asyncio.wait_for(self._log_event.wait_broadcast(), timeout=0.1)
                print(ev.content)
            except asyncio.TimeoutError:
                pass



"""StateManager to manage disks."""
import asyncio
import logging
from pathlib import Path
from typing import List

from astoria.common.manager import StateManager
from astoria.common.messages.astdiskd import (
    DiskInfo,
    DiskManagerMessage,
    DiskType,
)

from .disk_provider import DiskProvider
from .static import StaticDiskProvider
from .udisks import UdisksConnection

LOGGER = logging.getLogger(__name__)


class DiskManager(StateManager[DiskManagerMessage]):
    """Astoria Disk Manager."""

    name = "astdiskd"

    def _init(self) -> None:
        self._providers: List[DiskProvider] = [
            StaticDiskProvider(self, notify_coro=self.update_state),
        ]

        # Add UDisks provider if it DBus is installed
        if Path("/usr/bin/dbus-daemon").exists():
            self._providers.append(UdisksConnection(self, notify_coro=self.update_state))

    @property
    def offline_status(self) -> DiskManagerMessage:
        """
        Status to publish when the manager goes offline.

        This status should ensure that any other components relying
        on this data go into a safe state.
        """
        return DiskManagerMessage(
            status=DiskManagerMessage.Status.STOPPED,
            disks={},
        )

    async def main(self) -> None:
        """Main routine for astdiskd."""
        for provider in self._providers:
            asyncio.ensure_future(provider.main())

        # Wait whilst the program is running.
        await self.wait_loop()

    async def update_state(self) -> None:
        """Update the status of astdiskd when disks are changed."""
        disks = {}
        for provider in self._providers:
            for uuid, mount_path in provider.disks.items():
                disks[uuid] = DiskInfo(
                    uuid=uuid,
                    mount_path=mount_path,
                    disk_type=DiskType.determine_disk_type(mount_path),
                )
        self.status = DiskManagerMessage(
            status=DiskManagerMessage.Status.RUNNING,
            disks=disks,
        )



"""Command to list information about mounted disks."""
import asyncio
from typing import Optional

import click

from astoria.common.messages.astdiskd import DiskManagerMessage

from .command import SingleManagerMessageCommand

loop = asyncio.get_event_loop()


@click.command("list-disks")
@click.option("-v", "--verbose", is_flag=True)
@click.option("-c", "--config-file", type=click.Path(exists=True))
def list_disks(*, verbose: bool, config_file: Optional[str]) -> None:
    """List information about mounted disks."""
    command = ListDisksCommand(verbose, config_file)
    loop.run_until_complete(command.run())


class ListDisksCommand(SingleManagerMessageCommand[DiskManagerMessage]):
    """List disk information."""

    manager = "astdiskd"
    message_schema = DiskManagerMessage

    def handle_message(
        self,
        message: DiskManagerMessage,
    ) -> None:
        """Display information about the disks."""
        print(f"{len(message.disks)} disks found")
        for uuid, disk in message.disks.items():
            print(f"\tUUID: {uuid}")
            print(f"\t\tMounted at: {disk.mount_path}")
            print(f"\t\tDisk Type: {disk.disk_type.name}")



"""Command to show usercode info."""
import asyncio
from typing import Optional

import click

from astoria.common.messages.astprocd import ProcessManagerMessage
from astoria.consumers.astctl.command import SingleManagerMessageCommand

loop = asyncio.get_event_loop()


@click.command("show")
@click.option("-v", "--verbose", is_flag=True)
@click.option("-c", "--config-file", type=click.Path(exists=True))
def show(*, verbose: bool, config_file: Optional[str]) -> None:
    """Show current usercode."""
    command = ShowUsercodeCommand(verbose, config_file)
    loop.run_until_complete(command.run())


class ShowUsercodeCommand(SingleManagerMessageCommand[ProcessManagerMessage]):
    """Show current usercode."""

    manager = "astprocd"
    message_schema = ProcessManagerMessage

    def handle_message(
        self,
        message: ProcessManagerMessage,
    ) -> None:
        """Print the process status."""
        if message.code_status is not None and message.disk_info is not None:
            print(f"Code status: {message.code_status.value}")
            print(f"Disk Mountpoint: {message.disk_info.mount_path}")
        else:
            print("No usercode disk is inserted.")



"""MQTT message definitions for astdiskd."""
from enum import Enum
from pathlib import Path
from typing import Dict, NewType

from pydantic import BaseModel

from astoria.common.disk_constraints import (
    Constraint,
    FilePresentConstraint,
    TrueConstraint,
)

from .base import ManagerMessage

DiskUUID = NewType('DiskUUID', str)


class DiskType(Enum):
    """Type of disk."""

    USERCODE = "USERCODE"
    METADATA = "METADATA"
    UPDATE = "UPDATE"
    NOACTION = "NOACTION"

    @classmethod
    def determine_disk_type(cls, mount_path: Path) -> 'DiskType':
        """Determine the disk type from the mount path."""
        constraints: Dict['DiskType', Constraint] = {
            cls.USERCODE: FilePresentConstraint("robot.zip"),
            cls.METADATA: FilePresentConstraint("astoria.json"),
            cls.UPDATE: FilePresentConstraint("updatefile.txt"),
            cls.NOACTION: TrueConstraint(),  # Always match
        }

        for typ, constraint in constraints.items():
            if constraint.matches(mount_path):
                return typ

        raise RuntimeError("Unable to determine type of disk.")  # pragma: nocover


class DiskInfo(BaseModel):
    """Information about a mounted disk."""

    uuid: DiskUUID
    mount_path: Path
    disk_type: DiskType


class DiskManagerMessage(ManagerMessage):
    """
    Status message for Disk Manager.

    Published to /astoria/astdiskd
    """

    disks: Dict[DiskUUID, DiskInfo]



"""Disk Handler Mixin."""

import asyncio
import logging
from json import JSONDecodeError, loads
from typing import Dict, Match

from astoria.common.messages.astdiskd import (
    DiskInfo,
    DiskManagerMessage,
    DiskUUID,
)

LOGGER = logging.getLogger(__name__)


class DiskHandlerMixin:
    """Mixin to translate disk events into insertions and removals."""

    _cur_disks: Dict[DiskUUID, DiskInfo]

    async def handle_astdiskd_disk_info_message(
        self,
        match: Match[str],
        payload: str,
    ) -> None:
        """Handle disk info messages."""
        if payload:
            try:
                message = DiskManagerMessage(**loads(payload))

                new_set = set(message.disks.keys())
                old_set = set(self._cur_disks.keys())

                added_disks = new_set - old_set
                removed_disks = old_set - new_set

                for uuid in removed_disks:
                    info = self._cur_disks.pop(uuid)
                    asyncio.ensure_future(self.handle_disk_removal(uuid, info))

                for uuid in added_disks:
                    info = message.disks[uuid]
                    self._cur_disks[uuid] = info
                    asyncio.ensure_future(self.handle_disk_insertion(uuid, info))
            except JSONDecodeError:
                LOGGER.warning("Received bad JSON in disk manager message.")
        else:
            LOGGER.warning("Received empty disk manager message.")

    async def handle_disk_insertion(self, uuid: DiskUUID, disk_info: DiskInfo) -> None:
        """Handle a disk insertion."""
        LOGGER.debug(f"Disk inserted: {uuid} ({disk_info.disk_type})")  # pragma: nocover

    async def handle_disk_removal(self, uuid: DiskUUID, disk_info: DiskInfo) -> None:
        """Handle a disk removal."""
        LOGGER.debug(f"Disk removed: {uuid} ({disk_info.disk_type})")  # pragma: nocover



"""State Consumer base class."""
import asyncio
import logging
from abc import ABCMeta

from .data_component import DataComponent

LOGGER = logging.getLogger(__name__)

loop = asyncio.get_event_loop()


class StateConsumer(DataComponent, metaclass=ABCMeta):
    """
    State Consumer.

    A process that accesses some state.
    """

    async def _post_connect(self) -> None:
        """Overridable callback after MQTT connection."""
        await self._mqtt.wait_dependencies()



"""Command to set a metadata attribute."""
import asyncio
from typing import Optional

import click

from astoria.common.manager_requests import MetadataSetManagerRequest
from astoria.consumers.astctl.command import Command

loop = asyncio.get_event_loop()


@click.command("set")
@click.argument("attribute")
@click.argument("value")
@click.option("-v", "--verbose", is_flag=True)
@click.option("-c", "--config-file", type=click.Path(exists=True))
def set(attribute: str, value: str, *, verbose: bool, config_file: Optional[str]) -> None:
    """Set a metadata attribute."""
    command = SetMetadataCommand(attribute, value, verbose, config_file)
    loop.run_until_complete(command.run())


class SetMetadataCommand(Command):
    """Set a metadata attribute."""

    dependencies = ["astmetad"]

    def __init__(
        self,
        attribute: str,
        value: str,
        verbose: bool,
        config_file: Optional[str],
    ) -> None:
        super().__init__(verbose, config_file)
        self._attr = attribute
        self._value = value

    async def main(self) -> None:
        """Main method of the command."""
        res = await self._mqtt.manager_request(
            "astmetad",
            "mutate",
            MetadataSetManagerRequest(
                sender_name=self.name,
                attr=self._attr,
                value=self._value,
            ),
        )
        if res.success:
            print(f"Successfully set {self._attr} to {self._value}.")
            if len(res.reason) > 0:
                print(res.reason)
        else:
            print(f"Unable to set {self._attr} to {self._value}.")
            if len(res.reason) > 0:
                print(res.reason)
        # Add timeout
        self.halt(silent=True)



"""Manage running, killing and observing usercode."""

import asyncio
import logging
from datetime import datetime
from pathlib import Path
from signal import SIGKILL, SIGTERM
from tempfile import TemporaryDirectory
from typing import Callable, Dict, Optional, Set
from zipfile import BadZipFile, ZipFile

from astoria.common.broadcast_event import (
    LogEventSource,
    UsercodeLogBroadcastEvent,
)
from astoria.common.config import AstoriaConfig
from astoria.common.messages.astdiskd import DiskInfo, DiskUUID
from astoria.common.messages.astprocd import CodeStatus
from astoria.common.mqtt import BroadcastHelper

LOGGER = logging.getLogger(__name__)

loop = asyncio.get_event_loop()


class InvalidCodeBundleException(Exception):
    """The code bundle was not valid."""


class UsercodeLifecycle:
    """
    Manages the lifecycle of usercode.

    Handles process management and logging.
    """

    def __init__(
            self,
            uuid: DiskUUID,
            disk_info: DiskInfo,
            status_inform_callback: Callable[[CodeStatus], None],
            log_helper: BroadcastHelper[UsercodeLogBroadcastEvent],
            config: AstoriaConfig,
    ) -> None:
        self._uuid = uuid
        self._disk_info = disk_info
        self._status_inform_callback = status_inform_callback
        self._log_helper = log_helper
        self._config = config

        self._process: Optional[asyncio.subprocess.Process] = None
        self._process_lock = asyncio.Lock()
        self._setup_temp_dir()

        self.status = CodeStatus.STARTING

    @property
    def uuid(self) -> DiskUUID:
        """The UUID of the managed Disk."""
        return self._uuid

    @property
    def disk_info(self) -> DiskInfo:
        """Disk Info for the managed disk."""
        return self._disk_info

    @property
    def status(self) -> CodeStatus:
        """Get the status of the executing code."""
        return self._status

    @status.setter
    def status(self, status: CodeStatus) -> None:
        """Set the status of the executing code."""
        self._status = status
        self._status_inform_callback(status)

    def _setup_temp_dir(self) -> None:
        """Setup a temporary directory."""
        self._dir = TemporaryDirectory(prefix="astprocd-")
        self._dir_path = Path(self._dir.name)

    def _extract_and_validate_zip_file(self) -> None:
        """
        Extract and validate a robot.zip file.

        This function will extract and validate the contents of the
        robot.zip file. It will overwrite the current temporary dir
        of the usercode lifecycle, and should not be called multiple
        times.

        :raises InvalidCodeBundleException: The code bundle was not valid.
        """
        zip_path = self._disk_info.mount_path / "robot.zip"

        if not (zip_path.exists() and zip_path.is_file()):
            raise InvalidCodeBundleException("Unable to find robot.zip file")

        try:
            with ZipFile(zip_path, "r") as zf:
                zf.extractall(self._dir_path)
        except BadZipFile:
            raise InvalidCodeBundleException(
                "The provided robot.zip is not a valid ZIP archive",
            )

        # Check that robot.py is present
        exe_path = self._dir_path / "robot.py"
        if not exe_path.exists():
            raise InvalidCodeBundleException(
                "The provided robot.zip did not contain a robot.py",
            )

        LEGACY_FILES: Set[str] = {"info.yaml", "info.yml", "wifi.yaml", "wifi.yml"}

        for legacy_file in LEGACY_FILES:
            legacy_info_path = self._dir_path / legacy_file
            if legacy_info_path.exists():
                raise InvalidCodeBundleException(
                    "This is an old robot code package and \
                    will not work with this version of the kit.",
                )

    async def run_process(self) -> None:
        """
        Start the execution of the usercode.

        This function will not return until the code has exited.
        """
        if self._process is None:
            try:
                self._extract_and_validate_zip_file()
                async with self._process_lock:
                    self._process = await asyncio.create_subprocess_exec(
                        "python3",
                        "-u",
                        "robot.py",
                        stdin=asyncio.subprocess.DEVNULL,
                        stdout=asyncio.subprocess.PIPE,
                        stderr=asyncio.subprocess.PIPE,
                        cwd=self._dir_path,
                        start_new_session=True,
                    )
                    if self._process is not None:
                        if self._process.stdout is not None and \
                                self._process.stderr is not None:
                            asyncio.ensure_future(
                                self.logger(
                                    {LogEventSource.STDOUT: self._process.stdout,
                                     LogEventSource.STDERR: self._process.stderr},
                                ),
                            )
                        else:
                            LOGGER.warning("Unable to start logger task.")
                        self.status = CodeStatus.RUNNING
                        LOGGER.info(
                            f"Usercode pid {self._process.pid} "
                            f"started in {self._dir_path}",
                        )

                        # Wait for the subprocess to exit.
                        # This may include if it is killed.
                        rc = await self._process.wait()

                        if rc == 0:
                            self.status = CodeStatus.FINISHED
                        elif rc < 0:
                            self.status = CodeStatus.KILLED
                        elif rc > 0:
                            self.status = CodeStatus.CRASHED
                        LOGGER.info(
                            f"Usercode process exited with code {rc} "
                            f"({self.status.name})",
                        )

                        self._process = None

                        self._dir.cleanup()  # Reset directory
                        self._setup_temp_dir()
                    else:
                        LOGGER.warning("Tried to start process, but failed.")
                        self.status = CodeStatus.CRASHED  # Close enough to indicate error
            except InvalidCodeBundleException as e:
                LOGGER.warning("robot.zip was invalid. Unable to start code.")
                # Write error message to the log file.
                with self._disk_info.mount_path.joinpath("log.txt").open("w") as fh:
                    LOGGER.warning(str(e))
                    fh.write("Unable to start code.\n")
                    fh.write(f"{e}.\n")

                self.status = CodeStatus.CRASHED  # Close enough to indicate error
        else:
            LOGGER.warning("Tried to start process, but one is already running.")

    async def kill_process(self) -> None:
        """Kill the process, if one is running."""
        if self._process is not None and self._process_lock.locked():
            LOGGER.info("Attempting to kill process.")

            # Work-around for bpo-43884
            self._process._transport.close()  # type: ignore

            LOGGER.info(f"Sent SIGTERM to pid {self._process.pid}")
            self._process.send_signal(SIGTERM)
            try:
                await asyncio.wait_for(self._process.communicate(), timeout=5.0)
            except asyncio.TimeoutError:
                if self._process is not None:
                    LOGGER.info(f"Sent SIGKILL to pid {self._process.pid}")
                    self._process.send_signal(SIGKILL)
            except AttributeError:
                # Under some circumstances, there is a race condition such that
                # _process becomes None whilst the communicate timeout is running.
                # We want to catch and discard this error.
                pass
        else:
            LOGGER.debug("Tried to kill process, but no process is running.")

    async def logger(
        self,
        proc_outputs: Dict[LogEventSource, asyncio.StreamReader],
    ) -> None:
        """
        Logger task.

        Logs the output of the process to a log file and MQTT

        :param proc_outputs: streams of data from the usercode process
        """
        log_path = self._disk_info.mount_path / "log.txt"

        if self._process is not None:
            pid = self._process.pid
        else:
            pid = -1  # Use -1 if unknown

        def log(
            data: str,
            log_line_idx: int,
            source: LogEventSource = LogEventSource.ASTORIA,
        ) -> None:
            fh.write(data)
            fh.flush()
            self._log_helper.send(
                pid=pid,
                priority=log_line_idx,
                content=data,
                source=source,
            )

        with log_path.open("w") as fh:
            log_line = 0

            start_time = datetime.now()

            async def read_from_stream(
                    outputs: Dict[LogEventSource, asyncio.StreamReader],
                    source: LogEventSource,
                    log_line_idx: int,
            ) -> None:
                output = outputs[source]

                data = await output.readline()
                while data != b"":
                    data_str = data.decode('utf-8')
                    time_passed = datetime.now() - start_time
                    log(f"[{time_passed}] {data_str}", log_line_idx, source)
                    data = await output.readline()
                    log_line_idx += 1

            time_passed = datetime.now() - start_time

            # Print initial lines to the log, if any.
            # This is useful to show a message to the user in every log file.
            if self._config.system.initial_log_lines:
                log(f"[{time_passed}] ---\n", log_line)

                for line in self._config.system.initial_log_lines:
                    log(f"[{time_passed}] {line}\n", log_line)
                log(f"[{time_passed}] ---\n", log_line)

            log(f"[{time_passed}] === LOG STARTED ===\n", log_line)
            log_line += 1

            await asyncio.gather(
                read_from_stream(proc_outputs, LogEventSource.STDOUT, log_line),
                read_from_stream(proc_outputs, LogEventSource.STDERR, log_line),
            )
            time_passed = datetime.now() - start_time
            log(f"[{time_passed}] === LOG FINISHED ===\n", log_line)



"""Schema definitions for manager requests."""
from pathlib import Path
from uuid import UUID, uuid4

from pydantic import BaseModel, Field
from typing_extensions import final


class ManagerRequest(BaseModel):
    """Schema definition for a Manager Request."""

    uuid: UUID = Field(default_factory=uuid4)
    sender_name: str  # client name of component that sent the request


@final
class RequestResponse(BaseModel):
    """Schema definition for the response to a Manager Request."""

    uuid: UUID
    success: bool
    reason: str = ""


class MetadataSetManagerRequest(ManagerRequest):
    """Schema definition for a metadata mutation."""

    attr: str
    value: str


UsercodeKillManagerRequest = ManagerRequest
UsercodeRestartManagerRequest = ManagerRequest


class AddStaticDiskRequest(ManagerRequest):
    """Schema definition for adding a static disk."""

    path: Path


class RemoveStaticDiskRequest(ManagerRequest):
    """Schema definition for removing a static disk."""

    path: Path


RemoveAllStaticDisksRequest = ManagerRequest



"""Allows disks to be added manually through local filesystem paths."""
import logging
from pathlib import Path
from typing import TYPE_CHECKING, Callable, Coroutine, Dict

from astoria.common.manager_requests import (
    AddStaticDiskRequest,
    RemoveAllStaticDisksRequest,
    RemoveStaticDiskRequest,
    RequestResponse,
)
from astoria.common.messages.astdiskd import DiskUUID

from .disk_provider import DiskProvider

LOGGER = logging.getLogger(__name__)

if TYPE_CHECKING:
    from .disk_manager import DiskManager


class StaticDiskProvider(DiskProvider):
    """Provides disks added manually via a command."""

    def __init__(
        self,
        disk_manager: 'DiskManager',
        *,
        notify_coro: Callable[[], Coroutine[None, None, None]],
    ) -> None:
        super().__init__(disk_manager, notify_coro=notify_coro)
        self._disk_manager._register_request(
            "add_static_disk",
            AddStaticDiskRequest,
            self.handle_add_static_disk,
        )
        self._disk_manager._register_request(
            "remove_static_disk",
            RemoveStaticDiskRequest,
            self.handle_remove_static_disk,
        )
        self._disk_manager._register_request(
            "remove_all_static_disks",
            RemoveAllStaticDisksRequest,
            self.handle_remove_all_static_disks,
        )

    async def handle_add_static_disk(
        self,
        request: AddStaticDiskRequest,
    ) -> RequestResponse:
        """Handles the add static disk command."""
        if request.path.exists() and request.path.is_dir():
            if str(request.path) in map(lambda d: str(d), self.disks.values()):
                return RequestResponse(
                    uuid=request.uuid,
                    success=False,
                    reason="The specified path is already mounted.",
                )

            self.disks[DiskUUID(f"static-{request.uuid}")] = request.path
            await self._notify_coro()
            LOGGER.info(f"Static disk {request.uuid} mounted ({request.path})")
            return RequestResponse(
                uuid=request.uuid,
                success=True,
            )
        else:
            return RequestResponse(
                uuid=request.uuid,
                success=False,
                reason=f"{request.path} does not exist or is not a directory",
            )

    async def handle_remove_static_disk(
        self,
        request: RemoveStaticDiskRequest,
    ) -> RequestResponse:
        """Handles the remove static disk command."""
        if request.path in self.disks.values():
            for uuid, path in self.disks.items():
                if str(path) == str(request.path):
                    del self.disks[uuid]
                    LOGGER.info(f'Static disk {uuid} unmounted ({request.path})')
                    await self._notify_coro()
                    return RequestResponse(
                        uuid=request.uuid,
                        success=True,
                    )
            return RequestResponse(
                uuid=request.uuid,
                success=False,
                reason="Could not remove static disk.",
            )
        else:
            return RequestResponse(
                uuid=request.uuid,
                success=False,
                reason=f"{request.path} is not mounted as a static disk.",
            )

    async def handle_remove_all_static_disks(
        self,
        request: RemoveAllStaticDisksRequest,
    ) -> RequestResponse:
        """Handles the remove all static disks command."""
        # Find the disks that we need to remove
        removed_disks: Dict[DiskUUID, Path] = {
            uuid: path
            for uuid, path in self.disks.items()
            if uuid.startswith('static-')
        }

        # Remove those disks from the list of disks
        self._disks = {
            uuid: path
            for uuid, path in self.disks.items()
            if uuid not in removed_disks
        }

        # Log which disks we have remove
        for uuid, path in removed_disks.items():
            LOGGER.info(f'Static disk {uuid} unmounted ({path})')

        if len(removed_disks) == 0:
            return RequestResponse(
                uuid=request.uuid,
                success=True,
                reason='There are no static disks to remove.',
            )
        else:
            await self._notify_coro()
            return RequestResponse(
                uuid=request.uuid,
                success=True,
                reason='Successfully removed all static disks.',
            )



"""Metadata State Manager."""

import asyncio
import logging
from typing import Dict, List, Optional, Set, Tuple, Type

from astoria.common.manager import StateManager
from astoria.common.manager_requests import (
    MetadataSetManagerRequest,
    RequestResponse,
)
from astoria.common.messages.astdiskd import DiskInfo, DiskType, DiskUUID
from astoria.common.messages.astmetad import Metadata, MetadataManagerMessage
from astoria.managers.mixins.disk_handler import DiskHandlerMixin

from .metadata_cache import MetadataCache
from .metadata_disk_lifecycle import (
    AbstractMetadataDiskLifecycle,
    MetadataDiskLifecycle,
    UsercodeDiskLifecycle,
)

LOGGER = logging.getLogger(__name__)


class MetadataManager(DiskHandlerMixin, StateManager[MetadataManagerMessage]):
    """Astoria Metadata State Manager."""

    name = "astmetad"
    dependencies = ["astdiskd"]

    DISK_TYPE_LIFECYCLE_MAP: Dict[DiskType, Type[AbstractMetadataDiskLifecycle]] = {
        DiskType.USERCODE: UsercodeDiskLifecycle,
        DiskType.METADATA: MetadataDiskLifecycle,
    }

    DISK_TYPE_OVERRIDE_MAP: Dict[DiskType, Set[str]] = {
        DiskType.USERCODE: {"wifi_ssid", "wifi_psk", "wifi_region", "wifi_enabled"},
        DiskType.METADATA: {
            "arena", "zone", "mode", "marker_offset", "game_timeout", "wifi_enabled",
        },
    }

    MUTABLE_ATTRS_BY_REQUEST: Set[str] = {"arena", "zone", "mode"}
    CACHED_ATTRS: Set[str] = {"wifi_ssid", "wifi_psk", "wifi_region"}

    def _init(self) -> None:
        self._lifecycles: Dict[DiskType, Optional[AbstractMetadataDiskLifecycle]] = {
            disk_type: None
            for disk_type in self.DISK_TYPE_LIFECYCLE_MAP
        }
        self._cache = MetadataCache(
            self.CACHED_ATTRS,
            cache_path=self.config.system.cache_dir / "astmetad-metadata.json",
        )

        self._cur_disks: Dict[DiskUUID, DiskInfo] = {}
        self._mqtt.subscribe("astdiskd", self.handle_astdiskd_disk_info_message)

        self._requested_data: Dict[str, str] = {}
        self._register_request(
            "mutate",
            MetadataSetManagerRequest,
            self.handle_mutation_request,
        )

    @property
    def offline_status(self) -> MetadataManagerMessage:
        """
        Status to publish when the manager goes offline.

        This status should ensure that any other components relying
        on this data go into a safe state.
        """
        return MetadataManagerMessage(
            status=MetadataManagerMessage.Status.STOPPED,
            metadata=Metadata.init(self.config),
        )

    async def main(self) -> None:
        """Main routine for astmetad."""
        self.update_status()

        # Wait whilst the program is running.
        await self.wait_loop()

        for uuid, info in self._cur_disks.items():
            asyncio.ensure_future(self.handle_disk_removal(uuid, info))

    async def handle_disk_insertion(self, uuid: DiskUUID, disk_info: DiskInfo) -> None:
        """Handle a disk insertion."""
        LOGGER.debug(f"Disk inserted: {uuid} ({disk_info.disk_type})")
        for disk_type, lifecycle_class in self.DISK_TYPE_LIFECYCLE_MAP.items():
            if disk_info.disk_type is disk_type:
                LOGGER.info(
                    f"{disk_type.name} disk {uuid} is mounted"
                    f" at {disk_info.mount_path}",
                )
                if self._lifecycles[disk_type] is None:
                    LOGGER.debug(f"Starting lifecycle for {uuid}")
                    self._lifecycles[disk_type] = lifecycle_class(uuid, disk_info)
                    self.update_status()
                else:
                    LOGGER.warn(
                        "Cannot use metadata, there is already a lifecycle present.",
                    )

    async def handle_disk_removal(self, uuid: DiskUUID, disk_info: DiskInfo) -> None:
        """Handle a disk removal."""
        LOGGER.debug(f"Disk removed: {uuid} ({disk_info.disk_type})")
        for disk_type, lifecycle_class in self.DISK_TYPE_LIFECYCLE_MAP.items():
            if disk_info.disk_type is disk_type:
                LOGGER.info(f"Metadata disk {uuid} removed ({disk_info.mount_path})")
                lifecycle = self._lifecycles[disk_type]
                if lifecycle is not None and lifecycle._uuid == disk_info.uuid:
                    self._lifecycles[disk_type] = None
                    self.update_status()

    async def handle_mutation_request(
        self,
        request: MetadataSetManagerRequest,
    ) -> RequestResponse:
        """Handle a request to mutate metadata."""
        if request.attr not in self.MUTABLE_ATTRS_BY_REQUEST:
            return RequestResponse(
                uuid=request.uuid,
                success=False,
                reason=f"{request.attr} is not a mutable attribute",
            )

        if len(request.value) == 0:
            # Stop mutating the attr if it is empty.
            try:
                del self._requested_data[request.attr]
                LOGGER.info(f"{request.attr} override has been removed by request")
                self.update_status()
            except KeyError:
                pass
        else:
            self._requested_data[request.attr] = request.value
            LOGGER.info(
                f"{request.attr} has been overridden to {request.value} by request",
            )
            self.update_status()
        return RequestResponse(
            uuid=request.uuid,
            success=True,
        )

    def get_current_metadata(self) -> Metadata:
        """
        Calculate the current metadata.

        Takes the default, static metadata based on the config and system
        information. It then overlays data from other sources in a priority order,
        whereby each source has a set of permitted attributes in the metadata that
        can be overridden.
        """
        # Metadata sources in priority order.
        metadata_sources: List[Tuple[Set[str], Dict[str, str]]] = [
            (self.CACHED_ATTRS, self._cache.data),
            (self.MUTABLE_ATTRS_BY_REQUEST, self._requested_data),
        ]

        for disk_type, val in self._lifecycles.items():
            if val is not None:
                # Add disk-based metadata source if it is present.
                metadata_sources.append(
                    (
                        self.DISK_TYPE_OVERRIDE_MAP[disk_type],
                        val.diff_data,
                    ),
                )

        metadata = Metadata.init(self.config)

        for permitted_attrs, diff_data in metadata_sources:
            for k, v in diff_data.items():
                if k in permitted_attrs:
                    metadata.__setattr__(k, v)
                else:
                    LOGGER.warning(
                        f"There was an attempt to mutate {k}, but it was not permitted.",
                    )

        # Update the cache with the new values.
        for key in self.CACHED_ATTRS:
            self._cache.update_cached_attr(key, metadata.__getattribute__(key))

        return metadata

    def update_status(self) -> None:
        """Update the status of the manager."""
        self.status = MetadataManagerMessage(
            status=MetadataManagerMessage.Status.RUNNING,
            metadata=self.get_current_metadata(),
        )



"""Communicate with UDisks2 over DBus."""
import asyncio
import logging
from pathlib import Path
from typing import Dict, List

from dbus_next.aio import MessageBus
from dbus_next.aio.proxy_object import ProxyInterface
from dbus_next.constants import BusType
from dbus_next.errors import InterfaceNotFoundError
from dbus_next.signature import Variant

from astoria.common.messages.astdiskd import DiskUUID

from .disk_provider import DiskProvider

LOGGER = logging.getLogger(__name__)


class UdisksConnection(DiskProvider):
    """Connect and communicate with UDisks2."""

    DBUS_PATH: str = "/org/freedesktop/UDisks2"
    DBUS_NAME: str = "org.freedesktop.UDisks2"

    async def main(self) -> None:
        """Setup the message bus and task dispatcher."""
        mb = MessageBus(bus_type=BusType.SYSTEM)
        self._bus = await mb.connect()

        self._introspection = await self._bus.introspect(self.DBUS_NAME, self.DBUS_PATH)
        udisks_proxy = self._bus.get_proxy_object(
            self.DBUS_NAME,
            self.DBUS_PATH,
            self._introspection,
        )

        udisks_obj_manager = udisks_proxy.get_interface(
            "org.freedesktop.DBus.ObjectManager",
        )
        udisks_obj_manager.on_interfaces_added(self._disk_signal)

        await self._detect_initial_disks(udisks_obj_manager)

    def _bytes_to_path(self, data: List[int]) -> Path:
        """Convert a null terminated int array to a path."""
        # Data is null terminated.
        mount_point = data[:len(data) - 1]

        chars = [chr(x) for x in mount_point]
        mount_point_str = "".join(chars)

        return Path(mount_point_str)

    def _disk_signal(self, path: str, data: Dict[str, Dict[str, Variant]]) -> None:
        """
        Handle a disk signal event from UDisks2.

        Has to be synchronous due to limitations with python-dbus-next.

        Dispatches tasks to handle the events.
        """
        LOGGER.debug(f"disk_signal: Received disk signal from {path}")

        if path.startswith("/org/freedesktop/UDisks2/jobs/"):
            for job in data.keys():
                event_data = data[job]
                try:
                    if event_data["Operation"].value == "filesystem-mount":
                        LOGGER.debug(f"disk_signal: Mount Event detected at {path}")
                        if all((
                            'Objects' in event_data.keys(),
                            len(event_data["Objects"].value) > 0,
                        )):
                            disk_bus_path = event_data["Objects"].value[0]
                            LOGGER.debug(
                                f"disk_signal: Dispatching mount task"
                                f" for {disk_bus_path}",
                            )
                            asyncio.ensure_future(self.mount_task(disk_bus_path))
                        else:
                            LOGGER.warning(
                                f"No information available on disk at {path}, aborting.",
                            )

                    if event_data["Operation"].value == "cleanup":
                        LOGGER.debug(f"disk_signal: Removal Event detected at {path}")
                        asyncio.ensure_future(self.cleanup_task())
                except IndexError:
                    pass

    async def mount_task(self, disk_bus_path: str, *, notify: bool = True) -> None:
        """Handle a mount event."""
        await asyncio.sleep(0.3)  # Allow enough time for the mount to occur.

        introspection = await self._bus.introspect(self.DBUS_NAME, disk_bus_path)
        drive_obj = self._bus.get_proxy_object(
            self.DBUS_NAME,
            disk_bus_path,
            introspection,
        )

        try:
            drive_filesystem = drive_obj.get_interface(
                "org.freedesktop.UDisks2.Filesystem",
            )

            mount_points: List[List[int]] = \
                await drive_filesystem.get_mount_points()

            try:
                # We are only interested in the first mountpoint.
                mount_path = self._bytes_to_path(mount_points[0])
                if mount_path.exists() and mount_path.is_dir():
                    drive_block = drive_obj.get_interface("org.freedesktop.UDisks2.Block")
                    uuid: DiskUUID = DiskUUID(
                        await drive_block.get_id_uuid(),
                    )
                    if uuid not in self._disks.keys():
                        LOGGER.info(f"Disk {uuid} mounted ({mount_path})")
                        self._disks[uuid] = mount_path

                        if notify:
                            asyncio.ensure_future(self._notify_coro())
                    else:
                        LOGGER.error(f"Disk UUID collision! uuid={uuid}")
                else:
                    LOGGER.warning(f"Invalid mount path: {mount_path}")
            except IndexError:
                LOGGER.warning(f"No mount points available for disk at {disk_bus_path}")

        except InterfaceNotFoundError:
            # Object doesn't have a Filesystem interface
            pass

    async def cleanup_task(self, *, notify: bool = True) -> None:
        """Handle a cleanup event."""
        await asyncio.sleep(0.3)  # Allow enough time for the unmount to occur.

        # We have no information to tell which disk(s) left.
        # Thus we need to check all of them.
        removed_disks: List[DiskUUID] = []
        for uuid, path in self._disks.items():
            if not path.exists():
                LOGGER.info(f"Disks {uuid} removed ({path})")
                removed_disks.append(uuid)

        # We have to remove the disks here as we cannot modify
        # a dictionary as we are iterating over it.
        # Equally, we cannot break the previous loop in case a
        # disk had multiple partitions that were mounted.
        for uuid in removed_disks:
            self._disks.pop(uuid)

        if notify:
            asyncio.ensure_future(self._notify_coro())

    async def _detect_initial_disks(self, udisks_obj_manager: ProxyInterface) -> None:
        """Detect and register disks as startup."""
        LOGGER.info("Checking for initial disks at startup.")

        # The block devices are dbus objects managed by Udisks
        # We have to fetch them all unless we already know what they are.
        managed_objects: Dict[str, Dict[str, str]] = \
            await udisks_obj_manager.call_get_managed_objects()

        # Start a mount task for every block device and wait
        # for all of the tasks to be complete.
        tasks = (
            self.mount_task(path, notify=False)
            for path in managed_objects
            if path.startswith("/org/freedesktop/UDisks2/block_devices/")
        )
        await asyncio.gather(*tasks)

        # Send one notify
        asyncio.ensure_future(self._notify_coro())



"""Broadcast Event Schemas."""
from enum import Enum
from typing import ClassVar

from pydantic import BaseModel


class BroadcastEvent(BaseModel):
    """Schema for a broadcast event."""

    name: ClassVar[str]

    event_name: str
    sender_name: str
    priority: int = 0

    def __gt__(self, other: 'BroadcastEvent') -> bool:
        return self.priority > other.priority


class StartButtonBroadcastEvent(BroadcastEvent):
    """
    Schema for a remote start event.

    Trigger the robot code if it is waiting for the start button to be
    pressed. Does not affect or interact with the physical button as that
    is handled by the usercode driver.
    """

    name: ClassVar[str] = "start_button"


class LogEventSource(Enum):
    """The source of a line of log output."""

    ASTORIA = "astoria"
    STDOUT = "stdout"
    STDERR = "stderr"


class UsercodeLogBroadcastEvent(BroadcastEvent):
    """
    Schema for a log event from a usercode process.

    The priority of the event should be used to order the lines of the log.
    The pid should be used to differentiate between code runs.
    Content should be a single log line from the process, without a new line.
    """

    name: ClassVar[str] = "usercode_log"

    pid: int
    content: str
    source: LogEventSource



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['astoria',
 'astoria.common',
 'astoria.common.messages',
 'astoria.common.mqtt',
 'astoria.consumers',
 'astoria.consumers.astctl',
 'astoria.consumers.astctl.metadata',
 'astoria.consumers.astctl.static_disks',
 'astoria.consumers.astctl.usercode',
 'astoria.managers',
 'astoria.managers.astdiskd',
 'astoria.managers.astmetad',
 'astoria.managers.astprocd',
 'astoria.managers.mixins']

package_data = \
{'': ['*']}

install_requires = \
['click>=7.1.2,<8.0.0',
 'dbus-next>=0.2.1,<0.3.0',
 'gmqtt>=0.6.8,<0.7.0',
 'pydantic>=1.6.1,<2.0.0',
 'toml>=0.10.2,<0.11.0',
 'typing-extensions>=3.7.4,<4.0.0']

entry_points = \
{'console_scripts': ['astctl = astoria.consumers.astctl:main',
                     'astdiskd = astoria.managers.astdiskd:main',
                     'astmetad = astoria.managers.astmetad:main',
                     'astprocd = astoria.managers.astprocd:main',
                     'astwifid = astoria.consumers.astwifid:main']}

setup_kwargs = {
    'name': 'astoria',
    'version': '0.7.0',
    'description': 'Robot Management System for Student Robotics Kit',
    'long_description': '# Astoria\n\n![Tests](https://github.com/srobo/astoria/workflows/Tests/badge.svg)\n![Build and Publish docs](https://github.com/srobo/astoria/workflows/Build%20and%20Publish%20docs/badge.svg)\n[![PyPI version](https://badge.fury.io/py/astoria.svg)](https://badge.fury.io/py/astoria)\n[![MIT license](https://img.shields.io/badge/license-MIT-brightgreen.svg?style=flat)](https://opensource.org/licenses/MIT)\n![Bees](https://img.shields.io/badge/bees-110%25-yellow.svg)\n\nRobot Management System for Student Robotics 2022 kit.\n\nFor more information, please see the [documentation](https://srobo.github.io/astoria).\n\n## Contributions\n\nThis project is released under the MIT Licence. For more information, please see LICENSE.\n\nThe CONTRIBUTORS file can be generated by executing CONTRIBUTORS.gen. This generated file contains a list of people who have contributed to Astoria.\n\n',
    'author': 'Dan Trickey',
    'author_email': 'srobo-astoria@trickey.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/srobo/astoria',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



"""
USB Constraint.

Defines a set of parameters that a USB / Folder can conform to.
"""
from abc import ABCMeta, abstractmethod
from pathlib import Path


class Constraint(metaclass=ABCMeta):
    """A constraint that a path can match."""

    @abstractmethod
    def matches(self, path: Path) -> bool:
        """
        Determine if the disk at the given path matches the constraint.

        :param path: path to the mount point of the disk
        """
        raise NotImplementedError  # pragma: nocover


class FilePresentConstraint(Constraint):
    """
    Ensure that a file is present on the disk.

    This constraint will check that the path exists and contains the given file.
    """

    def __init__(self, filename: str) -> None:
        """
        Initialise the constraint class.

        :param filename: name of the file to find on the disk
        """
        self.filename = filename

    def matches(self, path: Path) -> bool:
        """
        Determine if the disk at the given path matches the constraint.

        :param path: path to the mount point of the disk
        """
        return all([
            path.exists(),
            path.is_dir(),
            path.joinpath(self.filename).exists(),
        ])

    def __repr__(self) -> str:
        return f"FilePresentConstraint(filename={self.filename})"


class NumberOfFilesConstraint(Constraint):
    """
    Ensure that a certain number of files are present.

    This can be used to ensure that there are no spurious files on the disk.
    """

    def __init__(self, n: int):
        """
        Initialise the constraint class.

        :param n: The number of files that should be on the disk.
        """
        self.n = n

    def matches(self, path: Path) -> bool:
        """
        Determine if the disk at the given path matches the constraint.

        :param path: path to the mount point of the disk
        """
        if all([
            path.exists(),
            path.is_dir(),
        ]):
            return self.n == len(list(path.iterdir()))
        else:
            return False

    def __repr__(self) -> str:
        return f"NumberOfFilesConstraint(n={self.n})"


class OrConstraint(Constraint):
    """Ensure that either of the given constraints match."""

    def __init__(self, a: Constraint, b: Constraint) -> None:
        """
        Initialise the constraint class.

        :param a: The first constraint
        :param b: The second constraint
        """
        self.a = a
        self.b = b

    def matches(self, path: Path) -> bool:
        """
        Determine if the disk at the given path matches the constraint.

        :param path: path to the mount point of the disk
        """
        return any([
            self.a.matches(path),
            self.b.matches(path),
        ])

    def __repr__(self) -> str:
        return f"OrConstraint(a={self.a}, b={self.b})"


class AndConstraint(Constraint):
    """Ensure that both of the constraints match."""

    def __init__(self, a: Constraint, b: Constraint) -> None:
        """
        Initialise the constraint class.

        :param a: The first constraint
        :param b: The second constraint
        """
        self.a = a
        self.b = b

    def matches(self, path: Path) -> bool:
        """
        Determine if the disk at the given path matches the constraint.

        :param path: path to the mount point of the disk
        """
        return all([
            self.a.matches(path),
            self.b.matches(path),
        ])

    def __repr__(self) -> str:
        return f"AndConstraint(a={self.a}, b={self.b})"


class NotConstraint(Constraint):
    """Ensure that the constraint does not match."""

    def __init__(self, a: Constraint):
        """
        Initialise the constraint class.

        :param a: The constraint to negate
        """
        self.a = a

    def matches(self, path: Path) -> bool:
        """
        Determine if the disk at the given path matches the constraint.

        :param path: path to the mount point of the disk
        """
        return not self.a.matches(path)

    def __repr__(self) -> str:
        return f"NotConstraint(a={self.a})"


class TrueConstraint(Constraint):
    """
    A constraint that is always true.

    Useful to create a default value when matching disks in order.
    """

    def matches(self, _: Path) -> bool:
        """
        Determine if the disk at the given path matches the constraint.

        :param _: path to the mount point of the disk. Not used.
        """
        return True

    def __repr__(self) -> str:
        return "TrueConstraint()"


class FalseConstraint(Constraint):
    """A constraint that is always false."""

    def matches(self, _: Path) -> bool:
        """
        Determine if the disk at the given path matches the constraint.

        :param _: path to the mount point of the disk. Not used.
        """
        return False

    def __repr__(self) -> str:
        return "FalseConstraint()"



"""Command base for astctl."""
import asyncio
from abc import abstractmethod
from json import JSONDecodeError, loads
from typing import Generic, Match, Type, TypeVar
from uuid import uuid4

from astoria.common.consumer import StateConsumer
from astoria.common.messages.base import ManagerMessage

T = TypeVar("T", bound=ManagerMessage)

loop = asyncio.get_event_loop()


class Command(StateConsumer):
    """
    Command base class for astctl.

    Disables the welcome message from the logger.
    """

    @property
    def name(self) -> str:
        """
        MQTT client name of the data component.

        This should be unique, as clashes will cause unexpected disconnections.
        """
        return f"astctl-{uuid4()}"

    def _setup_logging(self, verbose: bool, *, welcome_message: bool = True) -> None:
        super()._setup_logging(verbose, welcome_message=False)


class SingleManagerMessageCommand(Command, Generic[T]):
    """
    A command that waits for the message from a single manager and does something with it.

    Useful for getting data and displaying it.
    """

    @property
    @abstractmethod
    def manager(self) -> str:
        """The manager to receive a message from."""
        raise NotImplementedError

    @property
    @abstractmethod
    def message_schema(self) -> Type[T]:
        """The schema of the message for the manager."""
        raise NotImplementedError

    @abstractmethod
    def handle_message(
        self,
        message: T,
    ) -> None:
        """Do something with the message."""
        raise NotImplementedError

    def _init(self) -> None:
        """Initialise consumer."""
        self._received = False
        self._mqtt.subscribe(self.manager, self._handle_raw_message)

    async def main(self) -> None:
        """Main method of the command."""
        await self.wait_loop()

    async def _handle_raw_message(
        self,
        match: Match[str],
        payload: str,
    ) -> None:
        """Handle astdiskd status messages."""
        if not self._received:
            self._received = True
            try:
                message = self.message_schema(**loads(payload))
                if message.status == self.message_schema.Status.RUNNING:
                    self.handle_message(message)
                else:
                    print(f"{self.manager} is not running")
            except JSONDecodeError:
                print("Could not decode JSON data.")
        self.halt(silent=True)



"""
Astoria WiFi Daemon.

Manages a WiFi hotspot for the robot.
"""
import asyncio
import logging
import os
import signal
import tempfile
from json import JSONDecodeError, loads
from pathlib import Path
from typing import IO, Match, Optional

import click
from pydantic import ValidationError

from astoria.common.consumer import StateConsumer
from astoria.common.messages.astmetad import Metadata, MetadataManagerMessage

LOGGER = logging.getLogger(__name__)

loop = asyncio.get_event_loop()


@click.command("astwifid")
@click.option("-v", "--verbose", is_flag=True)
@click.option("-c", "--config-file", type=click.Path(exists=True))
def main(*, verbose: bool, config_file: Optional[str]) -> None:
    """The WiFi Daemon Application Entrypoint."""
    wifid = WiFiHotspotDaemon(verbose, config_file)
    loop.run_until_complete(wifid.run())


class WiFiHotspotDaemon(StateConsumer):
    """
    Hotspot management daemon.

    Receives metadata information from astmetad and manages the WiFi hotspot.
    """

    name = "astwifid"

    dependencies = ["astmetad"]

    def _init(self) -> None:
        self._lifecycle: Optional[WiFiHotspotLifeCycle] = None

        self._mqtt.subscribe("astmetad", self.handle_astmetad_message)

    async def main(self) -> None:
        """Main routine for astwifid."""
        # Wait whilst the program is running.
        await self.wait_loop()

        # Stop the hotspot when we shutdown
        if self._lifecycle:
            await self._lifecycle.stop_hotspot()

    async def handle_astmetad_message(
            self,
            match: Match[str],
            payload: str,
    ) -> None:
        """Event handler for metadata changes."""
        if payload:
            try:
                metadata_manager_message = MetadataManagerMessage(**loads(payload))
                await self.handle_metadata(metadata_manager_message.metadata)
            except ValidationError:
                LOGGER.warning("Received bad metadata manager message.")
            except JSONDecodeError:
                LOGGER.warning("Received bad JSON in metadata manager message.")
        else:
            LOGGER.warning("Received empty metadata manager message.")

    async def handle_metadata(self, metadata: Metadata) -> None:
        """
        Update the state of the hotspot based on the current metadata.

        :param metadata: The metadata included in the update.
        """
        wifi_interface = Path(f"/sys/class/net/{self.config.wifi.interface}")
        if self._lifecycle:
            if metadata.is_wifi_valid() and wifi_interface.exists():
                if self._lifecycle.has_metadata_changed(metadata):
                    await self._lifecycle.stop_hotspot()
                    self._lifecycle = WiFiHotspotLifeCycle(
                        # The types here are checked by is_wifi_valid
                        metadata.wifi_ssid,  # type: ignore
                        metadata.wifi_psk,  # type: ignore
                        metadata.wifi_region,  # type: ignore
                        self.config.wifi.interface,
                        self.config.wifi.bridge,
                        self.config.wifi.enable_wpa3,
                    )
                    asyncio.ensure_future(self._lifecycle.run_hotspot())
            else:
                # Turn it off!
                await self._lifecycle.stop_hotspot()
                self._lifecycle = None
        else:
            if metadata.is_wifi_valid() and wifi_interface.exists():
                # Turn it on!
                self._lifecycle = WiFiHotspotLifeCycle(
                    # The types here are checked by is_wifi_valid
                    metadata.wifi_ssid,  # type: ignore
                    metadata.wifi_psk,  # type: ignore
                    metadata.wifi_region,  # type: ignore
                    self.config.wifi.interface,
                    self.config.wifi.bridge,
                    self.config.wifi.enable_wpa3,
                )
                asyncio.ensure_future(self._lifecycle.run_hotspot())


class WiFiHotspotLifeCycle:
    """Manages the lifecycle of the hostapd process."""

    HOSTAPD_BINARY: str = "hostapd"

    def __init__(
            self,
            ssid: str,
            psk: str,
            region: str,
            interface: str,
            bridge: str,
            enable_wpa3: bool,
    ) -> None:
        LOGGER.info("Starting WiFi Hotspot lifecycle")
        self._ssid: str = ssid
        self._psk: str = psk
        self._region: str = region
        self._interface: str = interface
        self._bridge: str = bridge
        self._enable_wpa3: bool = enable_wpa3

        self._config_file: Optional[IO[bytes]] = None
        self._proc: Optional[asyncio.subprocess.Process] = None
        self._running: bool = False

    def has_metadata_changed(self, metadata: Metadata) -> bool:
        """Checks if the hotspot's properties match that of a set of metadata."""
        return not all(
            [
                self._ssid == metadata.wifi_ssid,
                self._psk == metadata.wifi_psk,
                self._region == metadata.wifi_region,
            ],
        )

    async def run_hotspot(self) -> None:
        """Starts the hostapd process."""
        self._running = True
        LOGGER.info(f"Starting WiFi Hotspot \"{self._ssid}\" on {self._interface}")
        self.generate_hostapd_config()
        if self._config_file is not None:
            while self._running:
                self._proc = await asyncio.create_subprocess_exec(
                    self.HOSTAPD_BINARY,
                    self._config_file.name,
                )
                LOGGER.info(f"{self.HOSTAPD_BINARY} started wth PID: {self._proc.pid}")
                sc = await self._proc.wait()
                if sc is None:
                    continue
                LOGGER.info(f"{self.HOSTAPD_BINARY} terminated with status code {sc}")

        else:
            raise RuntimeError(  # pragma: nocover
                "Tried to start hotspot, but the config file was not set.",
            )

    def generate_hostapd_config(self) -> None:
        """Generates a configuration file for hostapd based on the current metadata."""
        self._config_file = tempfile.NamedTemporaryFile(delete=False)
        LOGGER.debug(
            f"Writing {self.HOSTAPD_BINARY} configuration to {self._config_file.name}",
        )
        config = {
            "interface": self._interface,
            "bridge": self._bridge,
            "ssid": self._ssid,
            "country_code": self._region,
            "channel": 7,
            "hw_mode": "g",
            # Bit field: bit0 = WPA, bit1 = WPA2
            "wpa": 2,
            # Bit field: 1=wpa, 2=wep, 3=both
            "auth_algs": 1,
            # Set of accepted cipher suites; disabling insecure TKIP
            "wpa_pairwise": "CCMP",
            # Set of accepted key management algorithms
            # SAE = WPA3, WPA-PSK = WPA2
            "wpa_key_mgmt": "WPA-PSK",
            "wpa_passphrase": self._psk,
        }

        if self._enable_wpa3:
            config["wpa_key_mgmt"] = "SAE WPA-PSK"
            # Management frame support (802.11w)
            # Most client devices will not connect to a
            # WPA3-SAE secured AP unless it is using 802.11w.
            # This must be supported by both the AP and client.
            config["ieee80211w"] = 2
            config["sae_require_mfp"] = 1

        contents = "\n".join(f"{k}={v}" for k, v in config.items())

        self._config_file.write(contents.encode())
        self._config_file.close()

    async def stop_hotspot(self) -> None:
        """Stops the hostapd process."""
        self._running = False
        LOGGER.info("Stopping WiFi Hotspot")
        if self._proc is not None:
            self._proc.send_signal(signal.SIGINT)
            try:
                await asyncio.wait_for(self._proc.communicate(), timeout=5.0)
            except asyncio.TimeoutError:
                if self._proc is not None:
                    LOGGER.info(f"Sent SIGKILL to pid {self._proc.pid}")
                    self._proc.send_signal(signal.SIGKILL)
            except AttributeError:
                # Under some circumstances, there is a race condition such that
                # _proc becomes None whilst the communicate timeout is running.
                # We want to catch and discard this error.
                pass
            self._proc = None

        if self._config_file:
            os.unlink(self._config_file.name)


if __name__ == "__main__":
    main()



"""
Data Component base class.

A data component represents the common functionality between
State Managers and Consumers. It handles connecting to the broker
and managing the event loop.
"""
import asyncio
import logging
import signal
import sys
from abc import ABCMeta, abstractmethod
from signal import SIGHUP, SIGINT, SIGTERM
from types import FrameType
from typing import List, Optional

from pydantic import BaseModel

from astoria import __version__

from .config import AstoriaConfig
from .mqtt.wrapper import MQTTWrapper

LOGGER = logging.getLogger(__name__)

loop = asyncio.get_event_loop()


class DataComponent(metaclass=ABCMeta):
    """
    Data Component base class.

    A data component represents the common functionality between
    State Managers and Consumers. It handles connecting to the broker
    and managing the event loop.
    """

    config: AstoriaConfig

    def __init__(self, verbose: bool, config_file: Optional[str]) -> None:
        self.config = AstoriaConfig.load(config_file)

        self._setup_logging(verbose)
        self._setup_event_loop()
        self._setup_mqtt()

        self._init()

    def _setup_logging(self, verbose: bool, *, welcome_message: bool = True) -> None:
        if verbose:
            logging.basicConfig(
                level=logging.DEBUG,
                format=f"%(asctime)s {self.name} %(name)s %(levelname)s %(message)s",
                datefmt="%Y-%m-%d %H:%M:%S",
            )
        else:
            logging.basicConfig(
                level=logging.INFO,
                format=f"%(asctime)s {self.name} %(levelname)s %(message)s",
                datefmt="%Y-%m-%d %H:%M:%S",
            )

            # Suppress INFO messages from gmqtt
            logging.getLogger("gmqtt").setLevel(logging.WARNING)

        if welcome_message:
            LOGGER.info(f"{self.name} v{__version__} - {self.__doc__}")

    def _setup_event_loop(self) -> None:
        self._stop_event = asyncio.Event()

        loop.add_signal_handler(SIGHUP, self.halt)
        loop.add_signal_handler(SIGINT, self.halt)
        loop.add_signal_handler(SIGTERM, self.halt)

    def _setup_mqtt(self) -> None:
        self._mqtt = MQTTWrapper(
            self.name,
            self.config.mqtt,
            last_will=self.last_will,
            dependencies=self.dependencies,
            no_dependency_event=self._stop_event,
        )

    def _init(self) -> None:
        """
        Initialisation of the data component.

        Called in the constructor of the parent class.
        """
        pass

    def _exit(self, signals: signal.Signals, frame_type: FrameType) -> None:
        sys.exit(0)

    @property
    @abstractmethod
    def name(self) -> str:
        """
        MQTT client name of the data component.

        This should be unique, as clashes will cause unexpected disconnections.
        """
        raise NotImplementedError

    @property
    def dependencies(self) -> List[str]:
        """State Managers to depend on."""
        return []

    @property
    def last_will(self) -> Optional[BaseModel]:
        """Last will and testament of the MQTT client."""
        return None

    async def run(self) -> None:
        """Entrypoint for the data component."""
        await self._pre_connect()
        await self._mqtt.connect()
        await self._post_connect()

        await self.main()

        await self._pre_disconnect()
        await self._mqtt.disconnect()
        await self._post_disconnect()

    async def wait_loop(self) -> None:
        """Wait until the data component is halted."""
        await self._stop_event.wait()

    def halt(self, *, silent: bool = False) -> None:
        """Stop the component."""
        if not silent:
            LOGGER.info("Halting")
        self._stop_event.set()

    @abstractmethod
    async def main(self) -> None:
        """Main method of the data component."""
        raise NotImplementedError

    async def _pre_connect(self) -> None:
        """Overridable callback before MQTT connection."""
        pass

    async def _pre_disconnect(self) -> None:
        """Overridable callback before MQTT disconnection."""
        pass

    async def _post_connect(self) -> None:
        """Overridable callback after MQTT connection."""
        pass

    async def _post_disconnect(self) -> None:
        """Overridable callback after MQTT disconnection."""
        pass
